from nops_metadata import MetaFetcher
from nops_metadata import METAMAP
import pytest
import boto3


@pytest.fixture
def aws_session() -> boto3.Session:
    return boto3.Session()


@pytest.fixture()
def metafetcher(aws_session) -> MetaFetcher:
    return MetaFetcher(session=aws_session)


def test_meta_fetcher_init(metafetcher: MetaFetcher):
    assert metafetcher


def test_metamap_naming():
    """Metamap naming must be strongly consistent"""
    for key, element in METAMAP.items():
        assert element is not None
        _, entity_name = key.split("_", 1)

        fetch_method = element["fetch_method"]
        assert isinstance(fetch_method, str)
        _, entity_from_fetch_method = fetch_method.split("_", 1)
        assert entity_name == entity_from_fetch_method


def test_meta_fetcher_region_list(metafetcher: MetaFetcher):
    assert metafetcher.metadata_regions(metadata_type="ec2_instances")


def test_meta_fetcher_region_list_wrong_name(metafetcher: MetaFetcher):
    assert not metafetcher.metadata_regions(metadata_type="unknown_servicecall")


@pytest.fixture
def ecs_clusters(metafetcher: MetaFetcher):
    resources = metafetcher.fetch(metadata_type="ecs_clusters", region_name="us-west-2")
    return [r for r in resources]


def test_meta_fetcher_pulling_listing(ecs_clusters):
    assert ecs_clusters
    assert isinstance(ecs_clusters, list)

    for element in ecs_clusters:
        assert isinstance(element, dict)

    assert all(ecs_clusters)


def test_meta_fetcher_metadata_types(metafetcher):
    """Confirm that all calls are working"""
    for metadata_type in metafetcher.metadata_types:
        assert all(metafetcher.fetch(metadata_type=metadata_type, region_name="eu-central-1"))


def test_meta_fetcher_schema_export(metafetcher: MetaFetcher):
    for metadata_type in metafetcher.metadata_types:
        spark_schema = metafetcher.schema(metadata_type=metadata_type)
        postgres_schema = metafetcher.postgres_schema(metadata_type=metadata_type)
        assert spark_schema
        assert postgres_schema


def test_meta_fetcher_pulling_resource_details_payload(metafetcher: MetaFetcher, ecs_clusters: list[dict]):
    for resource_details_type in metafetcher.metadata_subtypes(metadata_type="ecs_clusters"):
        for resources in ecs_clusters[:1]:
            resource_details = metafetcher.fetch_resources(
                metadata_subtype=resource_details_type,
                region_name="us-west-2",
                resources=resources,
            )
            assert resource_details
            assert isinstance(resource_details, list)

            for detail in resource_details:
                assert detail
                assert isinstance(detail, dict)


# def test_meta_fetcher_pulling_resource_details_schema(metafetcher: MetaFetcher, ecs_clusters: list[dict]):
#     for metadata_subtype in metafetcher.metadata_subtypes(metadata_type="ecs_clusters"):
#         schema = metafetcher.schema(metadata_subtype=metadata_subtype)
#         assert schema



from typing import Any
from typing import Iterator

import boto3
from pyrsistent import freeze
from pyrsistent import thaw

from .constants import METAMAP
from .constants import SINLE_REGION_SERVICES
from .schema import SparkAWSSchemaSerializer
from .utils import resource_listing

METAMAP_DETAILS = freeze(
    {
        "ecs_clusters_details": {
            "fetch_method": "describe_clusters",
            "parameter_name": "clusters",
            "response_key": "clusters",
            "metadata_type": "ecs_clusters",
            "listing_key": "clusterArns",
            "kwargs": {"include": ["ATTACHMENTS", "CONFIGURATIONS", "SETTINGS", "STATISTICS", "TAGS"]},
        }
    }
)


class MetaFetcher:
    def __init__(self, session: boto3.Session):
        self.session: boto3.Session = session

    def metadata_regions(self, metadata_type: str) -> list[str]:
        service_name = metadata_type.split("_")[0]
        if service_name in SINLE_REGION_SERVICES:
            return ["us-east-1"]
        else:
            return self.session.get_available_regions(service_name=service_name)

    def schema(self, metadata_type: str) -> dict[str, Any]:
        serializer = SparkAWSSchemaSerializer()
        schema = serializer.schema(metadata_type)
        return schema

    def postgres_schema(self, metadata_type: str) -> dict[str, Any]:
        postgres_custom_types = {
            "string": "varchar",
            "long": "integer",
            "double": "float",
        }
        serializer = SparkAWSSchemaSerializer(custom_types=postgres_custom_types)
        schema = serializer.schema(metadata_type)
        return schema

    @property
    def metadata_types(self) -> list[str]:
        return list(METAMAP.keys())

    def metadata_subtypes(self, metadata_type: str) -> list[str]:
        response = []
        for key, value in METAMAP_DETAILS.items():
            if value["metadata_type"] == metadata_type:
                response.append(key)

        return response

    def fetch(self, metadata_type: str, region_name: str) -> Iterator[dict[str, Any]]:
        return resource_listing(
            session=self.session,
            metaname=metadata_type,
            fetch_method=METAMAP[metadata_type]["fetch_method"],
            response_key=METAMAP[metadata_type]["response_key"],
            page_key=METAMAP[metadata_type].get("page_key", ""),
            call_kwargs=thaw(METAMAP[metadata_type].get("kwargs", {})),
            region_name=region_name,
        )

    def fetch_resources(self, metadata_subtype: str, region_name: str, resources: dict) -> list[dict[str, Any]]:
        fetch_method: str = METAMAP_DETAILS[metadata_subtype]["fetch_method"]  # type: ignore
        parameter_name = METAMAP_DETAILS[metadata_subtype]["parameter_name"]
        response_key = METAMAP_DETAILS[metadata_subtype]["response_key"]
        call_kwargs = thaw(METAMAP_DETAILS[metadata_subtype].get("kwargs", {}))
        listing_key = METAMAP_DETAILS[metadata_subtype]["listing_key"]

        service = metadata_subtype.split("_")[0]
        client_kwargs = {}
        if region_name:
            client_kwargs["region_name"] = region_name

        client = self.session.client(service, **client_kwargs)

        if listing_key:
            resource_kwargs = {parameter_name: resources.get(listing_key)}
        else:
            resource_kwargs = {parameter_name: resources}

        response = getattr(client, fetch_method)(**call_kwargs, **resource_kwargs)[response_key]
        return response
